@@ -1001,6 +1001,26 @@ def draw_text(stream, textbox, offset_x, text_overflow, block_ellipsis):
     if textbox.style['visibility'] != 'visible':
         return
 
+    text_decoration_values = textbox.style['text_decoration_line']
+    text_decoration_color = textbox.style['text_decoration_color']
+    if text_decoration_color == 'currentColor':
+        text_decoration_color = textbox.style['color']
+    if 'overline' in text_decoration_values:
+        thickness = textbox.pango_layout.underline_thickness
+        offset_y = (
+            textbox.baseline - textbox.pango_layout.ascent + thickness / 2)
+        draw_text_decoration(
+            stream, textbox, offset_x, offset_y, thickness,
+            text_decoration_color)
+    if 'underline' in text_decoration_values:
+        thickness = textbox.pango_layout.underline_thickness
+        offset_y = (
+            textbox.baseline - textbox.pango_layout.underline_position +
+            thickness / 2)
+        draw_text_decoration(
+            stream, textbox, offset_x, offset_y, thickness,
+            text_decoration_color)
+
     x, y = textbox.position_x, textbox.position_y + textbox.baseline
     stream.set_color_rgb(*textbox.style['color'][:3])
     stream.set_alpha(textbox.style['color'][3])
@@ -1013,30 +1033,13 @@ def draw_text(stream, textbox, offset_x, text_overflow, block_ellipsis):
 
     draw_emojis(stream, textbox.style['font_size'], x, y, emojis)
 
-    # Draw text decoration
-    values = textbox.style['text_decoration_line']
-    color = textbox.style['text_decoration_color']
-    if color == 'currentColor':
-        color = textbox.style['color']
-    if 'overline' in values:
-        thickness = textbox.pango_layout.underline_thickness
-        offset_y = (
-            textbox.baseline - textbox.pango_layout.ascent + thickness / 2)
-        draw_text_decoration(
-            stream, textbox, offset_x, offset_y, thickness, color)
-    if 'underline' in values:
-        thickness = textbox.pango_layout.underline_thickness
-        offset_y = (
-            textbox.baseline - textbox.pango_layout.underline_position +
-            thickness / 2)
-        draw_text_decoration(
-            stream, textbox, offset_x, offset_y, thickness, color)
-    if 'line-through' in values:
+    if 'line-through' in text_decoration_values:
         thickness = textbox.pango_layout.strikethrough_thickness
         offset_y = (
             textbox.baseline - textbox.pango_layout.strikethrough_position)
         draw_text_decoration(
-            stream, textbox, offset_x, offset_y, thickness, color)
+            stream, textbox, offset_x, offset_y, thickness,
+            text_decoration_color)
 
     textbox.pango_layout.deactivate()
 
@@ -1052,6 +1055,10 @@ def draw_emojis(stream, font_size, x, y, emojis):
 def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
                     angle=0):
     """Draw the given ``textbox`` line to the document ``stream``."""
+    # Don’t draw lines with only invisible characters
+    if not textbox.text.strip():
+        return []
+
     font_size = textbox.style['font_size']
     if font_size < 1e-6:  # Default float precision used by pydyf
         return []
@@ -1198,8 +1205,7 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
                     png_data = ffi.unpack(hb_data, int(stream.length[0]))
                     pillow_image = Image.open(BytesIO(png_data))
                     image_id = f'{font.hash}{glyph}'
-                    image = RasterImage(
-                        pillow_image, image_id, optimize_size=())
+                    image = RasterImage(pillow_image, image_id, png_data)
                     d = font.widths[glyph] / 1000
                     a = pillow_image.width / pillow_image.height * d
                     pango.pango_font_get_glyph_extents(
@@ -1210,7 +1216,7 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
                     f = f / font_size - font_size
                     emojis.append([image, font, a, d, x_advance, f])
 
-            x_advance += (font.widths[glyph] + offset) / 1000
+            x_advance += (font.widths[glyph] + offset - kerning) / 1000
 
         # Close the last glyphs list, remove if empty
         if string[-1] == '<':
