@@ -7,8 +7,17 @@ are images with intrinsic dimensions.
 
 """
 
+try:
+    # Available in Python 3.9+
+    from importlib.resources import files
+except ImportError:
+    # Deprecated in Python 3.11+
+    from importlib.resources import read_text
+else:
+    def read_text(package, resource):
+        return (files(package) / resource).read_text('utf-8')
+
 import re
-from importlib.resources import read_text
 
 from . import CSS, css
 from .css import get_child_text
@@ -20,12 +29,15 @@ from .urls import get_url_attribute
 
 HTML5_UA_COUNTER_STYLE = CounterStyle()
 HTML5_UA = read_text(css, 'html5_ua.css')
+HTML5_UA_FORM = read_text(css, 'html5_ua_form.css')
 HTML5_PH = read_text(css, 'html5_ph.css')
 HTML5_UA_STYLESHEET = CSS(
     string=HTML5_UA, counter_style=HTML5_UA_COUNTER_STYLE)
+HTML5_UA_FORM_STYLESHEET = CSS(
+    string=HTML5_UA_FORM, counter_style=HTML5_UA_COUNTER_STYLE)
 HTML5_PH_STYLESHEET = CSS(string=HTML5_PH)
 
-# http://whatwg.org/C#space-character
+# https://html.spec.whatwg.org/multipage/#space-character
 HTML_WHITESPACE = ' \t\n\f\r'
 HTML_SPACE_SEPARATED_TOKENS_RE = re.compile(f'[^{HTML_WHITESPACE}]+')
 
@@ -33,20 +45,17 @@ HTML_SPACE_SEPARATED_TOKENS_RE = re.compile(f'[^{HTML_WHITESPACE}]+')
 def ascii_lower(string):
     r"""Transform (only) ASCII letters to lower case: A-Z is mapped to a-z.
 
-    :param string: An Unicode string.
-    :returns: A new Unicode string.
-
     This is used for `ASCII case-insensitive
-    <http://whatwg.org/C#ascii-case-insensitive>`_ matching.
+    <https://whatwg.org/C#ascii-case-insensitive>`_ matching.
 
     This is different from the :meth:`str.lower` method of Unicode strings
     which also affect non-ASCII characters,
     sometimes mapping them into the ASCII range:
 
-    >>> keyword = u'Bac\N{KELVIN SIGN}ground'
-    >>> assert keyword.lower() == u'background'
+    >>> keyword = 'Bac\N{KELVIN SIGN}ground'
+    >>> assert keyword.lower() == 'background'
     >>> assert ascii_lower(keyword) != keyword.lower()
-    >>> assert ascii_lower(keyword) == u'bac\N{KELVIN SIGN}ground'
+    >>> assert ascii_lower(keyword) == 'bac\N{KELVIN SIGN}ground'
 
     """
     # This turns out to be faster than unicode.translate()
@@ -54,15 +63,9 @@ def ascii_lower(string):
 
 
 def element_has_link_type(element, link_type):
-    """
-    Return whether the given element has a ``rel`` attribute with the
-    given link type.
-
-    :param link_type: Must be a lower-case string.
-
-    """
-    return any(ascii_lower(token) == link_type for token in
-               HTML_SPACE_SEPARATED_TOKENS_RE.findall(element.get('rel', '')))
+    """Return whether element has a ``rel`` attribute with given link type."""
+    tokens = HTML_SPACE_SEPARATED_TOKENS_RE.findall(element.get('rel', ''))
+    return any(ascii_lower(token) == link_type for token in tokens)
 
 
 # Maps HTML tag names to function taking an HTML element and returning a Box.
@@ -114,13 +117,14 @@ def handle_img(element, box, get_image_from_uri, base_url):
 
     Return either an image or the alt-text.
 
-    See: http://www.w3.org/TR/html5/embedded-content-1.html#the-img-element
+    See: https://www.w3.org/TR/html5/embedded-content-1.html#the-img-element
 
     """
     src = get_url_attribute(element, 'src', base_url)
     alt = element.get('alt')
     if src:
-        image = get_image_from_uri(url=src)
+        image = get_image_from_uri(
+            url=src, orientation=box.style['image_orientation'])
         if image is not None:
             return [make_replaced_box(element, box, image)]
         else:
@@ -154,7 +158,9 @@ def handle_embed(element, box, get_image_from_uri, base_url):
     src = get_url_attribute(element, 'src', base_url)
     type_ = element.get('type', '').strip()
     if src:
-        image = get_image_from_uri(url=src, forced_mime_type=type_)
+        image = get_image_from_uri(
+            url=src, forced_mime_type=type_,
+            orientation=box.style['image_orientation'])
         if image is not None:
             return [make_replaced_box(element, box, image)]
     # No fallback.
@@ -171,7 +177,9 @@ def handle_object(element, box, get_image_from_uri, base_url):
     data = get_url_attribute(element, 'data', base_url)
     type_ = element.get('type', '').strip()
     if data:
-        image = get_image_from_uri(url=data, forced_mime_type=type_)
+        image = get_image_from_uri(
+            url=data, forced_mime_type=type_,
+            orientation=box.style['image_orientation'])
         if image is not None:
             return [make_replaced_box(element, box, image)]
     # The element’s children are the fallback.
@@ -194,40 +202,11 @@ def handle_col(element, box, _get_image_from_uri, _base_url):
     """Handle the ``span`` attribute."""
     if isinstance(box, boxes.TableColumnBox) and box.span > 1:
         # Generate multiple boxes
-        # http://lists.w3.org/Archives/Public/www-style/2011Nov/0293.html
+        # https://lists.w3.org/Archives/Public/www-style/2011Nov/0293.html
         return [box.copy() for _i in range(box.span)]
     return [box]
 
 
-@handler('th')
-@handler('td')
-def handle_td(element, box, _get_image_from_uri, _base_url):
-    """Handle the ``colspan``, ``rowspan`` attributes."""
-    if isinstance(box, boxes.TableCellBox):
-        # HTML 4.01 gives special meaning to colspan=0
-        # http://www.w3.org/TR/html401/struct/tables.html#adef-rowspan
-        # but HTML 5 removed it
-        # http://www.w3.org/TR/html5/tabular-data.html#attr-tdth-colspan
-        # rowspan=0 is still there though.
-        try:
-            box.colspan = max(int(element.get('colspan', '').strip()), 1)
-        except ValueError:
-            box.colspan = 1
-        try:
-            box.rowspan = max(int(element.get('rowspan', '').strip()), 0)
-        except ValueError:
-            box.rowspan = 1
-    return [box]
-
-
-@handler('a')
-def handle_a(element, box, _get_image_from_uri, base_url):
-    """Handle the ``rel`` attribute."""
-    box.is_attachment = element_has_link_type(element, 'attachment')
-    box.download_name = element.get('download')
-    return [box]
-
-
 @handler('{http://www.w3.org/2000/svg}svg')
 def handle_svg(element, box, get_image_from_uri, base_url):
     """Handle ``<svg>`` elements.
@@ -252,10 +231,10 @@ def get_html_metadata(html):
 
     Relevant specs:
 
-    http://www.whatwg.org/html#the-title-element
-    http://www.whatwg.org/html#standard-metadata-names
-    http://wiki.whatwg.org/wiki/MetaExtensions
-    http://microformats.org/wiki/existing-rel-values#HTML5_link_type_extensions
+    https://www.whatwg.org/html#the-title-element
+    https://www.whatwg.org/html#standard-metadata-names
+    https://wiki.whatwg.org/wiki/MetaExtensions
+    https://microformats.org/wiki/existing-rel-values#HTML5_link_type_extensions
 
     """
     title = None
@@ -267,6 +246,7 @@ def get_html_metadata(html):
     modified = None
     attachments = []
     custom = {}
+    lang = html.etree_element.attrib.get('lang', None)
     for element in html.wrapper_element.query_all('title', 'meta', 'link'):
         element = element.etree_element
         if element.tag == 'title' and title is None:
@@ -305,15 +285,15 @@ def get_html_metadata(html):
     return dict(title=title, description=description, generator=generator,
                 keywords=keywords, authors=authors,
                 created=created, modified=modified,
-                attachments=attachments, custom=custom)
+                attachments=attachments, lang=lang, custom=custom)
 
 
 def strip_whitespace(string):
     """Use the HTML definition of "space character",
     not all Unicode Whitespace.
 
-    http://www.whatwg.org/html#strip-leading-and-trailing-whitespace
-    http://www.whatwg.org/html#space-character
+    https://www.whatwg.org/html#strip-leading-and-trailing-whitespace
+    https://www.whatwg.org/html#space-character
 
     """
     return string.strip(HTML_WHITESPACE)
