@@ -1,7 +1,6 @@
 """PDF stream."""
 
 import io
-import struct
 from functools import lru_cache
 from hashlib import md5
 
@@ -9,6 +8,7 @@ import pydyf
 from fontTools import subset
 from fontTools.ttLib import TTFont, TTLibError, ttFont
 from fontTools.varLib.mutator import instantiateVariableFont
+from PIL import Image
 
 from ..logger import LOGGER
 from ..matrix import Matrix
@@ -84,9 +84,9 @@ class Font:
         self.svg = harfbuzz.hb_ot_color_has_svg(hb_face)
         self.stemv = 80
         self.stemh = 80
-        self.bbox = [0, 0, 0, 0]
         self.widths = {}
         self.cmap = {}
+        self.used_in_forms = False
 
         # Font flags
         self.flags = 2 ** (3 - 1)  # Symbolic, custom character set
@@ -98,7 +98,7 @@ class Font:
         if len(widths) > 1 and len(set(widths)) == 1:
             self.flags += 2 ** (1 - 1)  # FixedPitch
 
-    def clean(self, cmap):
+    def clean(self, cmap, hinting):
         if self.ttfont is None:
             return
 
@@ -107,7 +107,7 @@ class Font:
             optimized_font = io.BytesIO()
             options = subset.Options(
                 retain_gids=True, passthrough_tables=True,
-                ignore_missing_glyphs=True, hinting=False,
+                ignore_missing_glyphs=True, hinting=hinting,
                 desubroutinize=True)
             options.drop_tables += ['GSUB', 'GPOS', 'SVG']
             subsetter = subset.Subsetter(options)
@@ -196,7 +196,6 @@ class Stream(pydyf.Stream):
     def __init__(self, fonts, page_rectangle, states, x_objects, patterns,
                  shadings, images, mark, *args, **kwargs):
         super().__init__(*args, **kwargs)
-        self.compress = True
         self.page_rectangle = page_rectangle
         self.marked = []
         self._fonts = fonts
@@ -297,7 +296,7 @@ class Stream(pydyf.Stream):
                 super().set_state(key)
 
     def set_alpha_state(self, x, y, width, height):
-        alpha_stream = self.add_group((x, y, width, height))
+        alpha_stream = self.add_group(x, y, width, height)
         alpha_state = pydyf.Dictionary({
             'Type': '/ExtGState',
             'SMask': pydyf.Dictionary({
@@ -319,18 +318,19 @@ class Stream(pydyf.Stream):
 
     @lru_cache()
     def add_font(self, pango_font):
-        if pango.pango_version() > 14600:
-            pango_face = pango.pango_font_get_face(pango_font)
-            description = pango.pango_font_face_describe(pango_face)
-        else:
-            description = pango.pango_font_describe(pango_font)
+        description = pango.pango_font_describe(pango_font)
+        mask = (
+            pango.PANGO_FONT_MASK_SIZE +
+            pango.PANGO_FONT_MASK_GRAVITY +
+            pango.PANGO_FONT_MASK_VARIATIONS)
+        pango.pango_font_description_unset_fields(description, mask)
         key = pango.pango_font_description_hash(description)
         pango.pango_font_description_free(description)
         if key not in self._fonts:
             self._fonts[key] = Font(pango_font)
         return self._fonts[key]
 
-    def add_group(self, bounding_box):
+    def add_group(self, x, y, width, height):
         states = pydyf.Dictionary()
         x_objects = pydyf.Dictionary()
         patterns = pydyf.Dictionary()
@@ -345,7 +345,7 @@ class Stream(pydyf.Stream):
         extra = pydyf.Dictionary({
             'Type': '/XObject',
             'Subtype': '/Form',
-            'BBox': pydyf.Array(bounding_box),
+            'BBox': pydyf.Array((x, y, x + width, y + height)),
             'Resources': resources,
             'Group': pydyf.Dictionary({
                 'Type': '/Group',
@@ -356,117 +356,36 @@ class Stream(pydyf.Stream):
         })
         group = Stream(
             self._fonts, self.page_rectangle, states, x_objects, patterns,
-            shadings, self._images, self._mark, extra=extra)
+            shadings, self._images, self._mark, extra=extra,
+            compress=self.compress)
         group.id = f'x{len(self._x_objects)}'
         self._x_objects[group.id] = group
         return group
 
-    def _get_png_data(self, pillow_image, optimize):
-        image_file = io.BytesIO()
-        pillow_image.save(image_file, format='PNG', optimize=optimize)
-
-        # Read the PNG header, then discard it because we know it's a PNG. If
-        # this weren't just output from Pillow, we should actually check it.
-        image_file.seek(8)
-
-        png_data = b''
-        raw_chunk_length = image_file.read(4)
-        # PNG files consist of a series of chunks.
-        while len(raw_chunk_length) > 0:
-            # Each chunk begins with its data length (four bytes, may be zero),
-            # then its type (four ASCII characters), then the data, then four
-            # bytes of a CRC.
-            chunk_len, = struct.unpack('!I', raw_chunk_length)
-            chunk_type = image_file.read(4)
-            if chunk_type == b'IDAT':
-                png_data += image_file.read(chunk_len)
-            else:
-                image_file.seek(chunk_len, io.SEEK_CUR)
-            # We aren't checking the CRC, we assume this is a valid PNG.
-            image_file.seek(4, io.SEEK_CUR)
-            raw_chunk_length = image_file.read(4)
-
-        return png_data
-
-    def add_image(self, pillow_image, image_rendering, optimize_size):
-        image_name = f'i{pillow_image.id}'
+    def add_image(self, image, width, height, interpolate, ratio):
+        image_name = f'i{image.id}{width}{height}{interpolate}{ratio}'
         self._x_objects[image_name] = None  # Set by write_pdf
         if image_name in self._images:
             # Reuse image already stored in document
             return image_name
 
-        if 'transparency' in pillow_image.info:
-            pillow_image = pillow_image.convert('RGBA')
-        elif pillow_image.mode in ('1', 'P', 'I'):
-            pillow_image = pillow_image.convert('RGB')
-
-        if pillow_image.mode in ('RGB', 'RGBA'):
-            color_space = '/DeviceRGB'
-        elif pillow_image.mode in ('L', 'LA'):
-            color_space = '/DeviceGray'
-        elif pillow_image.mode == 'CMYK':
-            color_space = '/DeviceCMYK'
-        else:
-            LOGGER.warning('Unknown image mode: %s', pillow_image.mode)
-            color_space = '/DeviceRGB'
-
-        interpolate = 'true' if image_rendering == 'auto' else 'false'
-        extra = pydyf.Dictionary({
-            'Type': '/XObject',
-            'Subtype': '/Image',
-            'Width': pillow_image.width,
-            'Height': pillow_image.height,
-            'ColorSpace': color_space,
-            'BitsPerComponent': 8,
-            'Interpolate': interpolate,
-        })
-
-        optimize = 'images' in optimize_size
-        if pillow_image.format == 'JPEG':
-            extra['Filter'] = '/DCTDecode'
+        if ratio != 1:
+            thumbnail = Image.open(io.BytesIO(image.image_data.data))
+            width = int(round(image.width * ratio))
+            height = int(round(image.height * ratio))
+            thumbnail.thumbnail((max(1, width), max(1, height)))
             image_file = io.BytesIO()
-            pillow_image.save(image_file, format='JPEG', optimize=optimize)
-            stream = [image_file.getvalue()]
-        else:
-            extra['Filter'] = '/FlateDecode'
-            extra['DecodeParms'] = pydyf.Dictionary({
-                # Predictor 15 specifies that we're providing PNG data,
-                # ostensibly using an "optimum predictor", but doesn't actually
-                # matter as long as the predictor value is 10+ according to the
-                # spec. (Other PNG predictor values assert that we're using
-                # specific predictors that we don't want to commit to, but
-                # "optimum" can vary.)
-                'Predictor': 15,
-                'Columns': pillow_image.width,
-            })
-            if pillow_image.mode in ('RGB', 'RGBA'):
-                # Defaults to 1.
-                extra['DecodeParms']['Colors'] = 3
-            if pillow_image.mode in ('RGBA', 'LA'):
-                alpha = pillow_image.getchannel('A')
-                pillow_image = pillow_image.convert(pillow_image.mode[:-1])
-                alpha_data = self._get_png_data(alpha, optimize)
-                extra['SMask'] = pydyf.Stream([alpha_data], extra={
-                    'Filter': '/FlateDecode',
-                    'Type': '/XObject',
-                    'Subtype': '/Image',
-                    'DecodeParms': pydyf.Dictionary({
-                        'Predictor': 15,
-                        'Columns': pillow_image.width,
-                    }),
-                    'Width': pillow_image.width,
-                    'Height': pillow_image.height,
-                    'ColorSpace': '/DeviceGray',
-                    'BitsPerComponent': 8,
-                    'Interpolate': interpolate,
-                    })
-            stream = [self._get_png_data(pillow_image, optimize)]
+            thumbnail.save(
+                image_file, format=thumbnail.format, optimize=image.optimize)
+            width, height = thumbnail.width, thumbnail.height
+            image.image_data = image.cache_image_data(image_file.getvalue())
 
-        xobject = pydyf.Stream(stream, extra=extra)
+        xobject = image.get_xobject(width, height, interpolate)
         self._images[image_name] = xobject
         return image_name
 
-    def add_pattern(self, width, height, repeat_width, repeat_height, matrix):
+    def add_pattern(self, x, y, width, height, repeat_width, repeat_height,
+                    matrix):
         states = pydyf.Dictionary()
         x_objects = pydyf.Dictionary()
         patterns = pydyf.Dictionary()
@@ -481,7 +400,7 @@ class Stream(pydyf.Stream):
         extra = pydyf.Dictionary({
             'Type': '/Pattern',
             'PatternType': 1,
-            'BBox': pydyf.Array([0, 0, width, height]),
+            'BBox': pydyf.Array([x, y, x + width, y + height]),
             'XStep': repeat_width,
             'YStep': repeat_height,
             'TilingType': 1,
@@ -491,7 +410,8 @@ class Stream(pydyf.Stream):
         })
         pattern = Stream(
             self._fonts, self.page_rectangle, states, x_objects, patterns,
-            shadings, self._images, self._mark, extra=extra)
+            shadings, self._images, self._mark, extra=extra,
+            compress=self.compress)
         pattern.id = f'p{len(self._patterns)}'
         self._patterns[pattern.id] = pattern
         return pattern
@@ -564,11 +484,7 @@ class Stream(pydyf.Stream):
             return element_tag.upper()
         elif element_tag in ('dl', 'ul', 'ol'):
             return 'L'
-        elif element_tag == 'li':
-            return 'LI'
-        elif element_tag == 'dt':
-            return 'LI'
-        elif element_tag == 'dd':
+        elif element_tag in ('li', 'dt', 'dd'):
             return 'LI'
         elif element_tag == 'table':
             return 'Table'
