@@ -1,6 +1,7 @@
 """Render SVG images."""
 
 import re
+from contextlib import suppress
 from math import cos, hypot, pi, radians, sin, sqrt
 from xml.etree import ElementTree
 
@@ -177,10 +178,16 @@ class Node:
     def get_href(self, base_url):
         """Get the href attribute, with or without a namespace."""
         for attr_name in ('{http://www.w3.org/1999/xlink}href', 'href'):
-            url = get_url_attribute(self, attr_name, base_url)
+            url = get_url_attribute(
+                self, attr_name, base_url, allow_relative=True)
             if url:
                 return url
 
+    def del_href(self):
+        """Remove the href attributes, with or without a namespace."""
+        for attr_name in ('{http://www.w3.org/1999/xlink}href', 'href'):
+            self.attrib.pop(attr_name, None)
+
     @staticmethod
     def process_whitespace(string, preserve):
         """Replace newlines by spaces, and merge spaces if not preserved."""
@@ -277,6 +284,13 @@ class Node:
             str(rotate.pop(0) if rotate else original_rotate[-1])
             for i in range(len(self.text)))
 
+    def override_iter(self, iterator):
+        """Override node’s children iterator."""
+        # As special methods are bound to classes and not instances, we have to
+        # create and assign a new type.
+        self.__class__ = type(
+            'Node', (Node,), {'__iter__': lambda _: iterator})
+
 
 class SVG:
     """An SVG document."""
@@ -287,6 +301,11 @@ class SVG:
         self.tree = Node(wrapper, style)
         self.url = url
 
+        # Replace 'currentColor' value
+        for key in COLOR_ATTRIBUTES:
+            if self.tree.get(key) == 'currentColor':
+                self.tree.attrib[key] = self.tree.get('color', 'black')
+
         self.filters = {}
         self.gradients = {}
         self.images = {}
@@ -302,6 +321,7 @@ class SVG:
         self.text_path_width = 0
 
         self.parse_defs(self.tree)
+        self.inherit_defs()
 
     def get_intrinsic_size(self, font_size):
         """Get intrinsic size of the image."""
@@ -383,7 +403,7 @@ class SVG:
                 coords = (box[0], box[1], box[0] + box[2], box[1] + box[3])
             else:
                 coords = (0, 0, self.inner_width, self.inner_height)
-            self.stream = self.stream.add_group(coords)
+            self.stream = self.stream.add_group(*coords)
 
         # Apply transform attribute
         self.transform(node.get('transform'), font_size)
@@ -415,10 +435,8 @@ class SVG:
 
         # Draw node
         if visible and node.tag in TAGS:
-            try:
+            with suppress(PointError):
                 TAGS[node.tag](self, node, font_size)
-            except PointError:
-                pass
 
         # Draw node children
         if display and node.tag not in DEF_TYPES:
@@ -706,6 +724,28 @@ class SVG:
         for child in node:
             self.parse_defs(child)
 
+    def inherit_defs(self):
+        """Handle inheritance of different defined elements lists."""
+        for defs in (self.gradients, self.patterns):
+            for element in defs.values():
+                self.inherit_element(element, defs)
+
+    def inherit_element(self, element, defs):
+        """Recursively handle inheritance of defined element."""
+        href = element.get_href(self.url)
+        if not href:
+            return
+        element.del_href()
+        parent = defs.get(parse_url(href).fragment)
+        if not parent:
+            return
+        self.inherit_element(parent, defs)
+        for key, value in parent.attrib.items():
+            if key not in element.attrib:
+                element.attrib[key] = value
+        if next(iter(element), None) is None:
+            element.override_iter(parent.__iter__())
+
     def calculate_bounding_box(self, node, font_size, stroke=True):
         """Calculate the bounding box of a node."""
         if stroke or node.bounding_box is None:
