@@ -5,11 +5,11 @@ from math import inf
 
 import pyphen
 
-from .constants import LST_TO_ISO, PANGO_STRETCH, PANGO_STYLE, PANGO_WRAP_MODE
+from .constants import LST_TO_ISO, PANGO_WRAP_MODE
 from .ffi import (
     ffi, gobject, pango, pangoft2, unicode_to_char_p, units_from_double,
     units_to_double)
-from .fonts import font_features
+from .fonts import font_features, get_font_description
 
 
 def line_size(line, style):
@@ -58,13 +58,13 @@ def first_line_metrics(first_line, text, layout, resume_at, space_collapse,
 
 class Layout:
     """Object holding PangoLayout-related cdata pointers."""
-    def __init__(self, context, font_size, style, justification_spacing=0,
+    def __init__(self, context, style, justification_spacing=0,
                  max_width=None):
         self.justification_spacing = justification_spacing
-        self.setup(context, font_size, style)
+        self.setup(context, style)
         self.max_width = max_width
 
-    def setup(self, context, font_size, style):
+    def setup(self, context, style):
         self.context = context
         self.style = style
         self.first_line_direction = 0
@@ -78,9 +78,6 @@ class Layout:
             pango.pango_font_map_create_context(font_map),
             gobject.g_object_unref)
         pango.pango_context_set_round_glyph_positions(pango_context, False)
-        self.layout = ffi.gc(
-            pango.pango_layout_new(pango_context),
-            gobject.g_object_unref)
 
         if style['font_language_override'] != 'normal':
             lang_p, lang = unicode_to_char_p(LST_TO_ISO.get(
@@ -97,26 +94,17 @@ class Layout:
 
         assert not isinstance(style['font_family'], str), (
             'font_family should be a list')
-        self.font = ffi.gc(
-            pango.pango_font_description_new(),
-            pango.pango_font_description_free)
-        family_p, family = unicode_to_char_p(','.join(style['font_family']))
-        pango.pango_font_description_set_family(self.font, family_p)
-        pango.pango_font_description_set_style(
-            self.font, PANGO_STYLE[style['font_style']])
-        pango.pango_font_description_set_stretch(
-            self.font, PANGO_STRETCH[style['font_stretch']])
-        pango.pango_font_description_set_weight(
-            self.font, style['font_weight'])
-        pango.pango_font_description_set_absolute_size(
-            self.font, units_from_double(font_size))
-        pango.pango_layout_set_font_description(self.layout, self.font)
+        font_description = get_font_description(style)
+        self.layout = ffi.gc(
+            pango.pango_layout_new(pango_context),
+            gobject.g_object_unref)
+        pango.pango_layout_set_font_description(self.layout, font_description)
 
         text_decoration = style['text_decoration_line']
         if text_decoration != 'none':
             metrics = ffi.gc(
                 pango.pango_context_get_metrics(
-                    pango_context, self.font, self.language),
+                    pango_context, font_description, self.language),
                 pango.pango_font_metrics_unref)
             self.ascent = units_to_double(
                 pango.pango_font_metrics_get_ascent(metrics))
@@ -141,7 +129,6 @@ class Layout:
         if features and context:
             features = ','.join(
                 f'{key} {value}' for key, value in features.items()).encode()
-            # TODO: attributes should be freed.
             # In the meantime, keep a cache to avoid leaking too many of them.
             attr = context.font_features.setdefault(
                 features, pango.pango_attr_font_features_new(features))
@@ -152,22 +139,17 @@ class Layout:
     def get_first_line(self):
         first_line = pango.pango_layout_get_line_readonly(self.layout, 0)
         second_line = pango.pango_layout_get_line_readonly(self.layout, 1)
-        if second_line != ffi.NULL:
-            index = second_line.start_index
-        else:
-            index = None
+        index = None if second_line == ffi.NULL else second_line.start_index
         self.first_line_direction = first_line.resolved_dir
         return first_line, index
 
     def set_text(self, text, justify=False):
-        try:
+        index = text.find('\n')
+        if index != -1:
             # Keep only the first line plus one character, we don't need more
-            text = text[:text.index('\n') + 2]
-        except ValueError:
-            # End-of-line not found, keep the whole text
-            pass
+            text = text[:index+2]
+        self.text = text
         text, bytestring = unicode_to_char_p(text)
-        self.text = bytestring.decode()
         pango.pango_layout_set_text(self.layout, text, -1)
 
         word_spacing = self.style['word_spacing']
@@ -185,12 +167,12 @@ class Layout:
 
         if self.text and (word_spacing or letter_spacing or word_breaking):
             attr_list = pango.pango_layout_get_attributes(self.layout)
-            if not attr_list:
-                # TODO: list should be freed
-                attr_list = pango.pango_attr_list_new()
+            if attr_list == ffi.NULL:
+                attr_list = ffi.gc(
+                    pango.pango_attr_list_new(),
+                    pango.pango_attr_list_unref)
 
             def add_attr(start, end, spacing):
-                # TODO: attributes should be freed
                 attr = pango.pango_attr_letter_spacing_new(spacing)
                 attr.start_index, attr.end_index = start, end
                 pango.pango_attr_list_change(attr_list, attr)
@@ -200,15 +182,23 @@ class Layout:
                 add_attr(0, len(bytestring), letter_spacing)
 
             if word_spacing:
+                if bytestring == b' ':
+                    # We need more than one space to set word spacing
+                    self.text = ' \u200b'  # Space + zero-width space
+                    text, bytestring = unicode_to_char_p(self.text)
+                    pango.pango_layout_set_text(self.layout, text, -1)
+
                 space_spacing = (
                     units_from_double(word_spacing) + letter_spacing)
                 position = bytestring.find(b' ')
+                # Pango gives only half of word-spacing on boundaries
+                boundary_positions = (0, len(bytestring) - 1)
                 while position != -1:
-                    add_attr(position, position + 1, space_spacing)
+                    factor = 1 + (position in boundary_positions)
+                    add_attr(position, position + 1, factor * space_spacing)
                     position = bytestring.find(b' ', position + 1)
 
             if word_breaking:
-                # TODO: attributes should be freed
                 attr = pango.pango_attr_insert_hyphens_new(False)
                 attr.start_index, attr.end_index = 0, len(bytestring)
                 pango.pango_attr_list_change(attr_list, attr)
@@ -222,8 +212,7 @@ class Layout:
     def set_tabs(self):
         if isinstance(self.style['tab_size'], int):
             layout = Layout(
-                self.context, self.style['font_size'], self.style,
-                self.justification_spacing)
+                self.context, self.style, self.justification_spacing)
             layout.set_text(' ' * self.style['tab_size'])
             line, _ = layout.get_first_line()
             width, _ = line_size(line, self.style)
@@ -238,25 +227,16 @@ class Layout:
         pango.pango_layout_set_tabs(self.layout, array)
 
     def deactivate(self):
-        del self.layout, self.font, self.language, self.style
+        del self.layout, self.language, self.style
 
     def reactivate(self, style):
-        self.setup(self.context, style['font_size'], style)
+        self.setup(self.context, style)
         self.set_text(self.text, justify=True)
 
 
 def create_layout(text, style, context, max_width, justification_spacing):
-    """Return an opaque Pango layout with default Pango line-breaks.
-
-    :param text: Unicode
-    :param style: a style dict of computed values
-    :param max_width:
-        The maximum available width in the same unit as ``style['font_size']``,
-        or ``None`` for unlimited width.
-
-    """
-    layout = Layout(
-        context, style['font_size'], style, justification_spacing, max_width)
+    """Return an opaque Pango layout with default Pango line-breaks."""
+    layout = Layout(context, style, justification_spacing, max_width)
 
     # Make sure that max_width * Pango.SCALE == max_width * 1024 fits in a
     # signed integer. Treat bigger values same as None: unconstrained width.
@@ -295,33 +275,29 @@ def split_first_line(text, style, context, max_width, justification_spacing,
         max_width = None
 
     # Step #1: Get a draft layout with the first line
-    layout = None
     if max_width is not None and max_width != inf and style['font_size']:
+        short_text = text
         if max_width == 0:
             # Trying to find minimum size, let's naively split on spaces and
             # keep one word + one letter
             space_index = text.find(' ')
-            if space_index == -1:
-                expected_length = len(text)
-            else:
-                expected_length = space_index + 2  # index + space + one letter
+            if space_index != -1:
+                short_text = text[:space_index+2]  # index + space + one letter
         else:
-            expected_length = int(max_width / style['font_size'] * 2.5)
-        if expected_length < len(text):
-            # Try to use a small amount of text instead of the whole text
-            layout = create_layout(
-                text[:expected_length], style, context, max_width,
-                justification_spacing)
-            first_line, index = layout.get_first_line()
-            if index is None:
-                # The small amount of text fits in one line, give up and use
-                # the whole text
-                layout = None
-    if layout is None:
+            short_text = text[:int(max_width / style['font_size'] * 2.5)]
+        # Try to use a small amount of text instead of the whole text
+        layout = create_layout(
+            short_text, style, context, max_width, justification_spacing)
+        first_line, resume_index = layout.get_first_line()
+        if resume_index is None and short_text != text:
+            # The small amount of text fits in one line, give up and use
+            # the whole text
+            layout.set_text(text)
+            first_line, resume_index = layout.get_first_line()
+    else:
         layout = create_layout(
             text, style, context, original_max_width, justification_spacing)
-        first_line, index = layout.get_first_line()
-    resume_index = index
+        first_line, resume_index = layout.get_first_line()
 
     # Step #2: Don't split lines when it's not needed
     if max_width is None:
@@ -329,7 +305,7 @@ def split_first_line(text, style, context, max_width, justification_spacing,
         return first_line_metrics(
             first_line, text, layout, resume_index, space_collapse, style)
     first_line_width, _ = line_size(first_line, style)
-    if index is None and first_line_width <= max_width:
+    if resume_index is None and first_line_width <= max_width:
         # The first line fits in the available width
         return first_line_metrics(
             first_line, text, layout, resume_index, space_collapse, style)
@@ -337,14 +313,14 @@ def split_first_line(text, style, context, max_width, justification_spacing,
     # Step #3: Try to put the first word of the second line on the first line
     # https://mail.gnome.org/archives/gtk-i18n-list/2013-September/msg00006
     # is a good thread related to this problem.
-    first_line_text = text.encode()[:index].decode()
+    first_line_text = text.encode()[:resume_index].decode()
     first_line_fits = (
         first_line_width <= max_width or
         ' ' in first_line_text.strip() or
         can_break_text(first_line_text.strip(), style['lang']))
     if first_line_fits:
         # The first line fits but may have been cut too early by Pango
-        second_line_text = text.encode()[index:].decode()
+        second_line_text = text.encode()[resume_index:].decode()
     else:
         # The line can't be split earlier, try to hyphenate the first word.
         first_line_text = ''
@@ -357,22 +333,19 @@ def split_first_line(text, style, context, max_width, justification_spacing,
             # only try when space collapsing is allowed
             new_first_line_text = first_line_text + next_word
             layout.set_text(new_first_line_text)
-            first_line, index = layout.get_first_line()
-            first_line_width, _ = line_size(first_line, style)
-            if index is None and first_line_text:
-                # The next word fits in the first line, keep the layout
-                resume_index = len(new_first_line_text.encode()) + 1
-                return first_line_metrics(
-                    first_line, text, layout, resume_index, space_collapse,
-                    style)
-            elif index:
-                # Text may have been split elsewhere by Pango earlier
-                resume_index = index
-            else:
-                # Second line is None
-                resume_index = first_line.length + 1
-                if resume_index >= len(text.encode()):
-                    resume_index = None
+            first_line, resume_index = layout.get_first_line()
+            if resume_index is None:
+                if first_line_text:
+                    # The next word fits in the first line, keep the layout
+                    resume_index = len(new_first_line_text.encode()) + 1
+                    return first_line_metrics(
+                        first_line, text, layout, resume_index, space_collapse,
+                        style)
+                else:
+                    # Second line is None
+                    resume_index = first_line.length + 1
+                    if resume_index >= len(text.encode()):
+                        resume_index = None
     elif first_line_text:
         # We found something on the first line but we did not find a word on
         # the next line, no need to hyphenate, we can keep the current layout
@@ -386,8 +359,10 @@ def split_first_line(text, style, context, max_width, justification_spacing,
     hyphenated = False
     soft_hyphen = '\xad'
 
-    try_hyphenate = False
+    auto_hyphenation = manual_hyphenation = False
     if hyphens != 'none':
+        manual_hyphenation = soft_hyphen in first_line_text + next_word
+    if hyphens == 'auto' and lang:
         next_word_boundaries = get_next_word_boundaries(second_line_text, lang)
         if next_word_boundaries:
             # We have a word to hyphenate
@@ -405,93 +380,81 @@ def split_first_line(text, style, context, max_width, justification_spacing,
                 if space > limit_zone or space < 0:
                     # Available space is worth the try, or the line is even too
                     # long to fit: try to hyphenate
-                    try_hyphenate = True
+                    auto_hyphenation = True
 
-    if try_hyphenate:
-        # Automatic hyphenation possible and next word is long enough
-        auto_hyphenation = hyphens == 'auto' and lang
-        manual_hyphenation = False
-        if auto_hyphenation:
-            if soft_hyphen in first_line_text or soft_hyphen in next_word:
-                # Automatic hyphenation opportunities within a word must be
-                # ignored if the word contains a conditional hyphen, in favor
-                # of the conditional hyphen(s).
-                # See https://drafts.csswg.org/css-text-3/#valdef-hyphens-auto
-                manual_hyphenation = True
-        else:
-            manual_hyphenation = hyphens == 'manual'
+    # Automatic hyphenation opportunities within a word must be ignored if the
+    # word contains a conditional hyphen, in favor of the conditional
+    # hyphen(s).
+    # See https://drafts.csswg.org/css-text-3/#valdef-hyphens-auto
+    if manual_hyphenation:
+        # Manual hyphenation: check that the line ends with a soft
+        # hyphen and add the missing hyphen
+        if first_line_text.endswith(soft_hyphen):
+            # The first line has been split on a soft hyphen
+            if ' ' in first_line_text:
+                first_line_text, next_word = first_line_text.rsplit(' ', 1)
+                next_word = f' {next_word}'
+                layout.set_text(first_line_text)
+                first_line, _ = layout.get_first_line()
+                resume_index = len((f'{first_line_text} ').encode())
+            else:
+                first_line_text, next_word = '', first_line_text
+        soft_hyphen_indexes = [
+            match.start() for match in re.finditer(soft_hyphen, next_word)]
+        soft_hyphen_indexes.reverse()
+        dictionary_iterations = [
+            next_word[:i + 1] for i in soft_hyphen_indexes]
+        start_word = 0
+    elif auto_hyphenation:
+        dictionary_key = (lang, left, right, total)
+        dictionary = context.dictionaries.get(dictionary_key)
+        if dictionary is None:
+            dictionary = pyphen.Pyphen(lang=lang, left=left, right=right)
+            context.dictionaries[dictionary_key] = dictionary
+        dictionary_iterations = [
+            start for start, end in dictionary.iterate(next_word)]
+    else:
+        dictionary_iterations = []
 
-        if manual_hyphenation:
-            # Manual hyphenation: check that the line ends with a soft
-            # hyphen and add the missing hyphen
-            if first_line_text.endswith(soft_hyphen):
-                # The first line has been split on a soft hyphen
-                if ' ' in first_line_text:
-                    first_line_text, next_word = (
-                        first_line_text.rsplit(' ', 1))
-                    next_word = f' {next_word}'
-                    layout.set_text(first_line_text)
-                    first_line, index = layout.get_first_line()
-                    resume_index = len((f'{first_line_text} ').encode())
-                else:
-                    first_line_text, next_word = '', first_line_text
-            soft_hyphen_indexes = [
-                match.start() for match in re.finditer(soft_hyphen, next_word)]
-            soft_hyphen_indexes.reverse()
-            dictionary_iterations = [
-                next_word[:i + 1] for i in soft_hyphen_indexes]
-        elif auto_hyphenation:
-            dictionary_key = (lang, left, right, total)
-            dictionary = context.dictionaries.get(dictionary_key)
-            if dictionary is None:
-                dictionary = pyphen.Pyphen(lang=lang, left=left, right=right)
-                context.dictionaries[dictionary_key] = dictionary
-            dictionary_iterations = [
-                start for start, end in dictionary.iterate(next_word)]
-        else:
-            dictionary_iterations = []
-
-        if dictionary_iterations:
-            for first_word_part in dictionary_iterations:
-                new_first_line_text = (
-                    first_line_text +
-                    second_line_text[:start_word] +
-                    first_word_part)
-                hyphenated_first_line_text = (
-                    new_first_line_text + style['hyphenate_character'])
-                new_layout = create_layout(
-                    hyphenated_first_line_text, style, context, max_width,
-                    justification_spacing)
-                new_first_line, new_index = new_layout.get_first_line()
-                new_first_line_width, _ = line_size(new_first_line, style)
-                new_space = max_width - new_first_line_width
-                if new_index is None and (
-                        new_space >= 0 or
-                        first_word_part == dictionary_iterations[-1]):
-                    hyphenated = True
-                    layout = new_layout
-                    first_line = new_first_line
-                    index = new_index
-                    resume_index = len(new_first_line_text.encode())
-                    if text[len(new_first_line_text)] == soft_hyphen:
-                        # Recreate the layout with no max_width to be sure that
-                        # we don't break before the soft hyphen
-                        pango.pango_layout_set_width(
-                            layout.layout, units_from_double(-1))
-                        resume_index += len(soft_hyphen.encode())
-                    break
-
-            if not hyphenated and not first_line_text:
-                # Recreate the layout with no max_width to be sure that
-                # we don't break before or inside the hyphenate character
-                hyphenated = True
-                layout.set_text(hyphenated_first_line_text)
-                pango.pango_layout_set_width(
-                    layout.layout, units_from_double(-1))
-                first_line, index = layout.get_first_line()
+    if dictionary_iterations:
+        for first_word_part in dictionary_iterations:
+            new_first_line_text = (
+                first_line_text +
+                second_line_text[:start_word] +
+                first_word_part)
+            hyphenated_first_line_text = (
+                new_first_line_text + style['hyphenate_character'])
+            new_layout = create_layout(
+                hyphenated_first_line_text, style, context, max_width,
+                justification_spacing)
+            new_first_line, index = new_layout.get_first_line()
+            new_first_line_width, _ = line_size(new_first_line, style)
+            new_space = max_width - new_first_line_width
+            hyphenated = (
+                index is None and (
+                    new_space >= 0 or
+                    first_word_part == dictionary_iterations[-1]))
+            if hyphenated:
+                layout = new_layout
+                first_line = new_first_line
                 resume_index = len(new_first_line_text.encode())
-                if text[len(first_line_text)] == soft_hyphen:
+                if text[len(new_first_line_text)] == soft_hyphen:
+                    # Recreate the layout with no max_width to be sure that
+                    # we don't break before the soft hyphen
+                    pango.pango_layout_set_width(layout.layout, -1)
                     resume_index += len(soft_hyphen.encode())
+                break
+
+        if not hyphenated and not first_line_text:
+            # Recreate the layout with no max_width to be sure that
+            # we don't break before or inside the hyphenate character
+            hyphenated = True
+            layout.set_text(hyphenated_first_line_text)
+            pango.pango_layout_set_width(layout.layout, -1)
+            first_line, _ = layout.get_first_line()
+            resume_index = len(new_first_line_text.encode())
+            if text[len(first_line_text)] == soft_hyphen:
+                resume_index += len(soft_hyphen.encode())
 
     if not hyphenated and first_line_text.endswith(soft_hyphen):
         # Recreate the layout with no max_width to be sure that
@@ -500,9 +463,8 @@ def split_first_line(text, style, context, max_width, justification_spacing,
         hyphenated_first_line_text = (
             first_line_text + style['hyphenate_character'])
         layout.set_text(hyphenated_first_line_text)
-        pango.pango_layout_set_width(
-            layout.layout, units_from_double(-1))
-        first_line, index = layout.get_first_line()
+        pango.pango_layout_set_width(layout.layout, -1)
+        first_line, _ = layout.get_first_line()
         resume_index = len(first_line_text.encode())
 
     # Step 5: Try to break word if it's too long for the line
